from typing import Any

import pytest
from pyemv import tlv


# fmt: off
@pytest.mark.parametrize(
    ["tags", "value"],
    [
        ("", {}),
        # Single byte length length
        ("9C00",       {"9C":     b""}),
        ("9C01FF",     {"9C":     b"\xFF"}),
        ("9C8101FF",   {"9C":     b"\xFF"}),
        ("9F0200",     {"9F02":   b""}),
        ("9F0201FF",   {"9F02":   b"\xFF"}),
        ("9F820100",   {"9F8201": b""}),
        ("9F820101FF", {"9F8201": b"\xFF"}),
        # Extended length
        ("9C80",           {"9C": b""}),
        ("9C8100",         {"9C": b""}),
        ("9C820000",       {"9C": b""}),
        ("9C83000000",     {"9C": b""}),
        ("9C8400000000",   {"9C": b""}),
        ("9C8101FF",       {"9C": b"\xFF"}),
        ("9C820001FF",     {"9C": b"\xFF"}),
        ("9C83000001FF",   {"9C": b"\xFF"}),
        ("9C8400000001FF", {"9C": b"\xFF"}),
    ],
)
# fmt: on
def test_decode_tag(tags: str, value: Any) -> None:
    r = tlv.decode(bytes.fromhex(tags))
    assert r == value


# fmt: off
@pytest.mark.parametrize(
    ["tags", "value"],
    [
        ("", {}),
        # Single byte length length
        ("9C00",       {"9C":     ""}),
        ("9C01FF",     {"9C":     "FF"}),
        ("9C8101FF",   {"9C":     "FF"}),
        ("9F0200",     {"9F02":   ""}),
        ("9F0201FF",   {"9F02":   "FF"}),
        ("9F820100",   {"9F8201": ""}),
        ("9F820101FF", {"9F8201": "FF"}),
        # Extended length
        ("9C80",           {"9C": ""}),
        ("9C8100",         {"9C": ""}),
        ("9C820000",       {"9C": ""}),
        ("9C83000000",     {"9C": ""}),
        ("9C8400000000",   {"9C": ""}),
        ("9C8101FF",       {"9C": "FF"}),
        ("9C820001FF",     {"9C": "FF"}),
        ("9C83000001FF",   {"9C": "FF"}),
        ("9C8400000001FF", {"9C": "FF"}),
    ],
)
# fmt: on
def test_decode_tag_coerce(tags: str, value: Any) -> None:
    r = tlv.decode(
        bytes.fromhex(tags), convert=lambda t, v: v.hex().upper(), flatten=True
    )
    assert r == value


def test_decode_nested() -> None:
    r = tlv.decode(b"\x9f\x02\x01\xff\xe0\x03\x9f\x03\x00\x9f\x04\x00")
    assert r == {"9F02": b"\xff", "E0": {"9F03": b""}, "9F04": b""}


def test_decode_nested_flat() -> None:
    r = tlv.decode(b"\x9f\x02\x01\xff\xe0\x03\x9f\x03\x00\x9f\x04\x00", flatten=True)
    assert r == {"9F02": b"\xff", "9F03": b"", "9F04": b""}

    r = tlv.decode(
        b"\x9f\x02\x01\xff\xe0\x04\x9f\x02\x01\xaa\x9f\x04\x00", flatten=True
    )
    assert r == {"9F02": b"\xaa", "9F04": b""}


def test_decode_simple() -> None:
    r = tlv.decode(b"\x9f\x02\xff" + b"\xee" * 255 + b"\x9f\x03\x00", simple=True)
    assert r == {"9F02": b"\xee" * 255, "9F03": b""}


# fmt: off
@pytest.mark.parametrize(
    ["tags", "tag", "offset", "value", "error"],
    [
        # Malformed simple tags (before recursion)
        # Missing tag
        ("9F",   "9F",   0, {}, "Tag malformed, expecting more data: tag '9F', offset 0."),
        ("9FF0", "9FF0", 0, {}, "Tag malformed, expecting more data: tag '9FF0', offset 0."),
        # Missing length
        ("9C",     "9C", 1, {}, "Tag length malformed, expecting 1 byte(s): tag '9C', offset 1."),
        ("9C81",   "9C", 2, {}, "Tag length malformed, expecting 1 byte(s): tag '9C', offset 2."),
        ("9C8200", "9C", 2, {}, "Tag length malformed, expecting 2 byte(s): tag '9C', offset 2."),
        # Missing data
        ("9C01",   "9C", 2, {}, "Tag value malformed, expecting 1 byte(s): tag '9C', offset 2."),
        ("9C0200", "9C", 2, {}, "Tag value malformed, expecting 2 byte(s): tag '9C', offset 2."),
        ("E001",   "E0", 2, {}, "Tag value malformed, expecting 1 byte(s): tag 'E0', offset 2."),

        # Malformed constructed tags
        # Data must be preset after constructed tag to make things more complex.
        # Reported error must be  accurate.
        # Missing tag
        ("E0019F0001FF",       "9F", 2, {'E0': {}}, "Tag malformed, expecting more data: tag '9F', offset 2."),
        # Missing length
        ("E0019C01FF",         "9C", 3, {'E0': {}}, "Tag length malformed, expecting 1 byte(s): tag '9C', offset 3."),
        ("E0029C8101FF",       "9C", 4, {'E0': {}}, "Tag length malformed, expecting 1 byte(s): tag '9C', offset 4."),
        ("E0039C8201FF",       "9C", 4, {'E0': {}}, "Tag length malformed, expecting 2 byte(s): tag '9C', offset 4."),
        # Missing data
        ("E0039C8101C001FF",   "9C", 5, {'E0': {}}, "Tag value malformed, expecting 1 byte(s): tag '9C', offset 5."),
        ("E0049C810200C001FF", "9C", 5, {'E0': {}}, "Tag value malformed, expecting 2 byte(s): tag '9C', offset 5."),

        # Make sure exception contains correct TLV data
        ("9C01FFAA",     "AA", 4, {"9C": b"\xFF"},         "Tag length malformed, expecting 1 byte(s): tag 'AA', offset 4."),
        ("E0039C01FFAA", "AA", 6, {"E0": {"9C": b"\xFF"}}, "Tag length malformed, expecting 1 byte(s): tag 'AA', offset 6."),
    ],
)
# fmt: on
def test_decode_exception(
    tags: str, tag: str, offset: int, value: Any, error: str
) -> None:
    tags_b = bytes.fromhex(tags)
    with pytest.raises(tlv.DecodeError) as e:
        _ = tlv.decode(tags_b)
    assert e.value.args[0] == error
    assert e.value.tag == tag
    assert e.value.offset == offset
    assert e.value.tlv == value


# fmt: off
@pytest.mark.parametrize(
    ["tags", "value"],
    [
        ({}, ""),
        # Single byte length length
        ({"9C": b""},         "9C00"),
        ({"9C": b"\xFF"},     "9C01FF"),
        ({"9F00": b"\xFF"},   "9F0001FF"),
        ({"9F8001": b"\xFF"}, "9F800101FF"),
        # Extended length
        ({"9C": b"\xFF"*128}, "9C8180"+"FF"*128),
        ({"9C": b"\xFF"*511}, "9C8201FF"+"FF"*511),
        # Constructed
        ({"E0": {}},                "E000"),
        ({"E0": {"9C": b""}},       "E0029C00"),
        ({"E0": {"9C": b"\xFF"}},   "E0039C01FF"),
        ({"E0": {"9F00": b"\xFF"}}, "E0049F0001FF"),
        ({"E0": {"9F00": b"\xFF"}, "9C": b"\xFF"}, "E0049F0001FF9C01FF"),
    ],
)
# fmt: on
def test_encode(tags: Any, value: str) -> None:
    r = tlv.encode(tags)
    assert r == bytes.fromhex(value)


# fmt: off
@pytest.mark.parametrize(
    ["tags", "value"],
    [
        ({}, ""),
        # Single byte length length
        ({"9C": ""},       "9C00"),
        ({"9C": "FF"},     "9C01FF"),
        ({"9F00": "FF"},   "9F0001FF"),
        ({"9F8001": "FF"}, "9F800101FF"),
        # Extended length
        ({"9C": "FF"*128}, "9C8180"+"FF"*128),
        ({"9C": "FF"*511}, "9C8201FF"+"FF"*511),
        # Constructed
        ({"E0": {}},                "E000"),
        ({"E0": {"9C": ""}},       "E0029C00"),
        ({"E0": {"9C": "FF"}},   "E0039C01FF"),
        ({"E0": {"9F00": "FF"}}, "E0049F0001FF"),
        ({"E0": {"9F00": "FF"}, "9C": "FF"}, "E0049F0001FF9C01FF"),
    ],
)
# fmt: on
def test_encode_str(tags: Any, value: str) -> None:
    r = tlv.encode(tags)
    assert r == bytes.fromhex(value)


# fmt: off
@pytest.mark.parametrize(
    ["tags", "value"],
    [
        ({}, ""),
        # Single byte length length
        ({"9C": b""},         "9C00"),
        ({"9C": b"\xFF"},     "9C01FF"),
        ({"9F00": b"\xFF"},   "9F0001FF"),
        ({"9F8001": b"\xFF"}, "9F800101FF"),
        # Extended length
        ({"9C": b"\xFF"*128}, "9C80"+"FF"*128),
        ({"9C": b"\xFF"*255}, "9CFF"+"FF"*255),
        # Constructed
        ({"E0": {}},                "E000"),
        ({"E0": {"9C": b""}},       "E0029C00"),
        ({"E0": {"9C": b"\xFF"}},   "E0039C01FF"),
        ({"E0": {"9F00": b"\xFF"}}, "E0049F0001FF"),
        ({"E0": {"9F00": b"\xFF"}, "9C": b"\xFF"}, "E0049F0001FF9C01FF"),
    ],
)
# fmt: on
def test_encode_simple(tags: Any, value: str) -> None:
    r = tlv.encode(tags, simple=True)
    assert r == bytes.fromhex(value)


# fmt: off
@pytest.mark.parametrize(
    ["tags", "tag", "error"],
    [
        # Tag
        ({"XX": b""},     "XX",     "Invalid tag format, expecting hexchar string: tag 'XX'."),
        ({"00X": b""},    "00X",    "Invalid tag format, expecting hexchar string: tag '00X'."),
        ({"0": b""},      "0",      "Invalid tag format, expecting hexchar string: tag '0'."),
        ({"9F": b""},     "9F",     "Invalid tag format, expecting more data: tag '9F'."),
        ({"9F0001": b""}, "9F0001", "Invalid tag format, extra data: tag '9F0001'."),
        # Value
        ({"EC": []},   "EC",  "Invalid value type (list) for a constructed tag, expecting a dict: tag 'EC'."),
        ({"EC": ()},   "EC",  "Invalid value type (tuple) for a constructed tag, expecting a dict: tag 'EC'."),
        ({"9C": []},   "9C",  "Invalid value type (list) for a primitive tag, expecting bytes or str: tag '9C'."),
        ({"9C": ()},   "9C",  "Invalid value type (tuple) for a primitive tag, expecting bytes or str: tag '9C'."),
        ({"9C": "0"},  "9C", "Invalid value format, expecting hexchar string: tag '9C'."),
        ({"9C": "X"},  "9C", "Invalid value format, expecting hexchar string: tag '9C'."),
    ],
)
# fmt: on
def test_encode_exception(tags: Any, tag: str, error: str) -> None:
    with pytest.raises(tlv.EncodeError) as e:
        _ = tlv.encode(tags)
    assert e.value.args[0] == error
    assert e.value.tag == tag


# fmt: off
@pytest.mark.parametrize(
    ["tags", "tag", "error"],
    [
        ({"9C": b"\xFF" * 256}, "9C", "Value length (256) cannot exceed 255 bytes when 'simple' is enabled: tag '9C'."),
        ({"9C": "FF" * 256},    "9C", "Value length (256) cannot exceed 255 bytes when 'simple' is enabled: tag '9C'."),
    ],
)
# fmt: on
def test_encode_simple_exception(tags: Any, tag: str, error: str) -> None:
    with pytest.raises(tlv.EncodeError) as e:
        _ = tlv.encode(tags, simple=True)
    assert e.value.args[0] == error
    assert e.value.tag == tag



import pytest

from pyemv import sm


def test_generate_command_mac_exception() -> None:
    # SK < 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        sm.generate_command_mac(
            sk_smi=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            command=bytes.fromhex("12345678901214"),
        )

    # SK > 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        sm.generate_command_mac(
            sk_smi=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            command=bytes.fromhex("12345678901214"),
        )


def test_encrypt_command_data_mastercard_nonpadded() -> None:
    # If MasterCard data is not multiple of 8 bytes
    # then it has to be padded with 0x80 and 0x00 until
    # it is multiple of 8 bytes.
    # if the data is multiple of 8 bytes then no padding
    # is required.

    no_padding_required = sm.encrypt_command_data(
        sk_smc=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
        command_data=bytes.fromhex("1234800000000000"),
        encryption_type=sm.EncryptionType.MASTERCARD,
    )

    padding_required = sm.encrypt_command_data(
        sk_smc=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
        command_data=bytes.fromhex("1234"),
        encryption_type=sm.EncryptionType.MASTERCARD,
    )

    assert no_padding_required == padding_required


def test_encrypt_command_data_exception() -> None:
    # SK < 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        sm.encrypt_command_data(
            sk_smc=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            command_data=bytes.fromhex("12345678901214"),
            encryption_type=sm.EncryptionType.VISA,
        )

    # SK > 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        sm.encrypt_command_data(
            sk_smc=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            command_data=bytes.fromhex("12345678901214"),
            encryption_type=sm.EncryptionType.VISA,
        )

    # Invalid encryption type
    with pytest.raises(
        TypeError,
        match="Encryption type must be EncryptionType Enum, not dict",
    ):
        sm.encrypt_command_data(
            sk_smc=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            command_data=bytes.fromhex("12345678901214"),
            encryption_type={},  # type: ignore
        )


def test_format_vis_pin_block_exception() -> None:
    # PIN < 4 bytes
    with pytest.raises(
        ValueError,
        match="PIN must be between 4 and 12 digits long",
    ):
        sm.format_vis_pin_block(
            icc_mk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            pin=b"123",
        )

    # PIN > 12 bytes
    with pytest.raises(
        ValueError,
        match="PIN must be between 4 and 12 digits long",
    ):
        sm.format_vis_pin_block(
            icc_mk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            pin=b"1234567890123",
        )

    # ICC MK AC < 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key for AC must be a double length DES key",
    ):
        sm.format_vis_pin_block(
            icc_mk_ac=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            pin=b"123456789012",
        )

    # ICC MK AC > 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key for AC must be a double length DES key",
    ):
        sm.format_vis_pin_block(
            icc_mk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            pin=b"123456789012",
        )

    # Current PIN < 4 bytes
    with pytest.raises(
        ValueError,
        match="PIN must be between 4 and 12 digits long",
    ):
        sm.format_vis_pin_block(
            icc_mk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            pin=b"1234",
            current_pin=b"123",
        )

    # Current PIN > 12 bytes
    with pytest.raises(
        ValueError,
        match="PIN must be between 4 and 12 digits long",
    ):
        sm.format_vis_pin_block(
            icc_mk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            pin=b"1234",
            current_pin=b"1234567890123",
        )


def test_format_iso9564_2_pin_block_exception() -> None:
    # PIN < 4 bytes
    with pytest.raises(
        ValueError,
        match="PIN must be between 4 and 12 digits long",
    ):
        sm.format_iso9564_2_pin_block(pin=b"123")

    # PIN > 12 bytes
    with pytest.raises(
        ValueError,
        match="PIN must be between 4 and 12 digits long",
    ):
        sm.format_iso9564_2_pin_block(pin=b"1234567890123")



r"""The objectives of secure messaging are to ensure data confidentiality, data
integrity, and authentication of the sender. Data integrity and issuer
authentication are achieved using a MAC. Data confidentiality is achieved
using encipherment of the data field.

Secure Messaging Integrity (MAC):

    >>> import pyemv
    >>> mac = pyemv.sm.generate_command_mac(
    ...     sk_smi=bytes.fromhex('0123456789ABCDEFFEDCBA9876543210'),
    ...     command=bytes.fromhex('8424000008'))
    >>> mac.hex().upper()
    '0BFFF5DF3FAA24E1'

Secure Messaging Confidentiality:

    >>> import pyemv
    >>> pin_block=pyemv.sm.format_iso9564_2_pin_block('9999')
    >>> encrypted_pin = pyemv.sm.encrypt_command_data(
    ...     sk_smc=bytes.fromhex('0123456789ABCDEFFEDCBA9876543210'),
    ...     command_data=pin_block,
    ...     encryption_type=pyemv.sm.EncryptionType.EMV)
    >>> encrypted_pin.hex().upper()
    '5A862D1381CCB94822CFDD706A376178'
"""

import binascii as _binascii
import sys as _sys
import typing as _typing
from enum import Enum as _Enum

from pyemv import mac as _mac
from pyemv import tools as _tools

__all__ = [
    "generate_command_mac",
    "EncryptionType",
    "encrypt_command_data",
    "format_vis_pin_block",
    "format_iso9564_2_pin_block",
]


def generate_command_mac(
    sk_smi: bytes, command: bytes, length: _typing.Optional[int] = None
) -> bytes:
    r"""Message Authentication Code (MAC) for Issuer Script Integrity.

    Parameters
    ----------
    sk_smi : bytes
        Binary ICC Session Key for script integrity (MAC).
    command : bytes
        Binary command to be MACed; may or may not include command data, e.g:

            Command Header || { Command Data if present }

        Some issuers choose to append transaction data, such as ATC and ARQC,
        to the command between header and data. The transaction data is present
        for the MAC calculation but not transmitted.

            Command Header || ATC || ARQC || { Command Data if present }

    length : int, optional
        Desired length of AC [4 <= `length` <= 8] (default 8 bytes).

    Returns
    -------
    mac : bytes
        Binary command MAC.

    Raises
    ------
    ValueError
        Session Key must be a double length DES key

    Notes
    -----
    During a transaction the host may send an Issuer Script. These
    Issuer Scripts allow the issuer to block or unblock an application,
    or unblock or change PIN on the card among other things. These issuer
    scripts are extracted from the host response and sent to the card as
    individual commands. These commands contain a MAC value that the card
    authenticates prior to actioning the script. The MAC can be 4-8 bytes
    in length.

    For further details see also:
        - EMV 4.3 Book 2 Section 9.2 Secure Messaging for Integrity and
          Authentication
        - EMV 4.3 Book 2 Section 9.2.1.2 Format 2
        - EMV 4.3 Book 2 Section 9.2.3 MAC Computation
        - EMV 4.3 Book 2 Annex A 1.2

    Examples
    --------
    >>> from pyemv import sm
    >>> sk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> command = bytes.fromhex("8424000008")
    >>> atc = bytes.fromhex("FFFF")
    >>> arqc = bytes.fromhex("1234567890123456")
    >>> mac = sm.generate_command_mac(sk_smi, command + atc + arqc)
    >>> mac.hex().upper()
    'E07B8DF1B4184282'
    """
    if len(sk_smi) != 16:
        raise ValueError("Session Key must be a double length DES key")

    return _mac.mac_iso9797_3(sk_smi[:8], sk_smi[-8:], command, 2, length)


class EncryptionType(_Enum):
    VISA = 1
    MASTERCARD = 2
    EMV = 3


def encrypt_command_data(
    sk_smc: bytes, command_data: bytes, encryption_type: EncryptionType
) -> bytes:
    r"""Command Data Encryption for Issuer Script Confidentiality.

    Parameters
    ----------
    sk_smc : bytes
        Binary ICC Session Key for script confidentiality.
    command_data : bytes
        Binary command data, e.g. PUT DATA or PIN block.
    encryption_type : EncryptionType
        Defines triple DES mode and padding method of `command_data`:

            - VISA = Prepend the command data with one byte containing the
              length of the command data. Then pad data according to
              ISO/IEC 9797-1 method 2.
            - MASTERCARD = If the command data is not multiple of 8 bytes then
              pad data according to ISO/IEC 9797-1 method 2.
            - EMV = Pad data according to ISO/IEC 9797-1 method 2.

    Returns
    -------
    encrypted_command_data : bytes
        Binary encrypted command data. Then the resulting command is:

            Header || Encrypted Data || MAC


    Raises
    ------
    ValueError
        Session Key must be a double length DES key
    TypeError
        Encryption type must be EncryptionType Enum

    Notes
    -----
    The Issuer Script sent by the host may contain command data, i.e.
    PUT DATA command on card data or the PIN Change issuer script.
    The issuer may decide to encrypt the command data

    For further details see also:
        - EMV 4.3 Book 2 Section 9.3 Secure Messaging for Confidentiality
        - EMV 4.3 Book 2 Section 9.3.1.2 Format 2
        - EMV 4.3 Book 2 Annex A 1.1

    See Also
    --------
    pyemv.mac.pad_iso9797_2 : ISO/IEC 9797-1 padding method 2

    Examples
    --------
    >>> from pyemv import sm
    >>> sk_smc = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> pin_block = bytes.fromhex("241234FFFFFFFFFF")
    >>> command_data = sm.encrypt_command_data(
    ...     sk_smc, pin_block, sm.EncryptionType.MASTERCARD)
    >>> command_data.hex().upper()
    '9859240AE52820C3'
    """
    if len(sk_smc) != 16:
        raise ValueError("Session Key must be a double length DES key")

    # Prepend data length as a single byte then
    # pad according to ISO/IEC 9797-1 method 2
    if encryption_type == EncryptionType.VISA:
        return _tools.encrypt_tdes_ecb(
            sk_smc,
            _mac.pad_iso9797_2(
                len(command_data).to_bytes(1, _sys.byteorder) + command_data, 8
            ),
        )

    # If the data is not multiple of 8 bytes then
    # pad according to ISO/IEC 9797-1 method 2
    if encryption_type == EncryptionType.MASTERCARD:
        if len(command_data) % 8 > 0:
            return _tools.encrypt_tdes_cbc(
                sk_smc,
                b"\x00\x00\x00\x00\x00\x00\x00\x00",
                _mac.pad_iso9797_2(command_data, 8),
            )
        return _tools.encrypt_tdes_cbc(
            sk_smc, b"\x00\x00\x00\x00\x00\x00\x00\x00", command_data
        )

    # Always pad according to ISO/IEC 9797-1 method 2
    if encryption_type == EncryptionType.EMV:
        return _tools.encrypt_tdes_cbc(
            sk_smc,
            b"\x00\x00\x00\x00\x00\x00\x00\x00",
            _mac.pad_iso9797_2(command_data, 8),
        )

    raise TypeError(
        "Encryption type must be EncryptionType Enum, "
        f"not {encryption_type.__class__.__name__}"
    )


def format_vis_pin_block(
    icc_mk_ac: bytes,
    pin: _typing.Union[bytes, str],
    current_pin: _typing.Optional[_typing.Union[bytes, str]] = None,
) -> bytes:
    r"""Format VIS PIN block with or without the current PIN.

    Parameters
    ----------
    icc_mk_ac : bytes
        Binary 16-byte ICC Master Key for Application Cryptogram
        Has to be a valid DES key.
    pin : bytes or str
        New ASCII Personal Identification Number.
    current_pin : bytes or str, optional
        Current ASCII Personal Identification Number (optional). If present
        VIS PIN block is generated using current PIN.

    Returns
    -------
    pin_block : bytes
        Binary 16-byte VIS PIN block

    Raises
    ------
    ValueError
        PIN must be between 4 and 12 digits long
    ValueError
        Current PIN must be between 4 and 12 digits long
    ValueError
        ICC Master Key for AC must be a double length DES key

    Examples
    --------
    >>> from pyemv import sm
    >>> icc_mk_ac = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> sm.format_vis_pin_block(icc_mk_ac, "1234").hex().upper()
    '041234FF76543210'
    """
    if len(pin) < 4 or len(pin) > 12:
        raise ValueError("PIN must be between 4 and 12 digits long")

    if len(icc_mk_ac) != 16:
        raise ValueError("ICC Master Key for AC must be a double length DES key")

    if isinstance(pin, bytes):
        pin = pin.decode("ascii")

    if isinstance(current_pin, bytes):
        current_pin = current_pin.decode("ascii")

    # 4 right-most bytes of ICC MK AC Key A padded with 0x00 to form an 8-byte block
    block_a = b"\x00" * 4 + icc_mk_ac[4:8]

    # PIN length as 1 byte concatenated with PIN then F-padded to form an 8 byte block
    block_b = len(pin).to_bytes(1, _sys.byteorder) + _binascii.a2b_hex(
        pin + "F" * (14 - len(pin))
    )

    pin_block = _tools.xor(block_a, block_b)

    # Generate VIS PIN block using current PIN
    if current_pin is not None:
        if len(current_pin) < 4 or len(current_pin) > 12:
            raise ValueError("Current PIN must be between 4 and 12 digits long")

        pin_block = _tools.xor(
            pin_block, _binascii.a2b_hex(current_pin + "0" * (16 - len(current_pin)))
        )

    return pin_block


def format_iso9564_2_pin_block(pin: _typing.Union[bytes, str]) -> bytes:
    r"""Format ISO 9564-1 PIN block format 2.

    Parameters
    ----------
    pin : bytes or str
        New ASCII Personal Identification Number.

    Returns
    -------
    pin_block : bytes
        Binary 8-byte PIN block

    Raises
    ------
    ValueError
        PIN must be between 4 and 12 digits long

    Examples
    --------
    >>> from pyemv import sm
    >>> sm.format_iso9564_2_pin_block("123456789012").hex().upper()
    '2C123456789012FF'
    """
    if len(pin) < 4 or len(pin) > 12:
        raise ValueError("PIN must be between 4 and 12 digits long")

    if isinstance(pin, bytes):
        pin = pin.decode("ascii")

    return (len(pin) + 32).to_bytes(1, _sys.byteorder) + _binascii.a2b_hex(
        pin + "F" * (14 - len(pin))
    )



from typing import Union

import pytest
from pyemv import kd, sm, tools


def test_generate_command_mac() -> None:
    """
    Test generation of script command data MAC

    Master Key Derivation = Option A
    SMI using Common Session Key Derivation (EMV SU-46)
    MAC generation script command is padded with 0x80
    """
    # Verify issuer master key check digits
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "1234567890123456"
    psn = "01"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 2).hex().upper() == "0239"

    # Verify SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("ABCDEF1234567890")
    sk_smi = kd.derive_common_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 2).hex().upper() == "F088"

    # Script MAC generated using Session Key
    script_command = bytes.fromhex("8424000008")
    mac = sm.generate_command_mac(sk_smi, script_command)
    assert mac.hex().upper() == "CF323F09A0F6AB9E"


@pytest.mark.parametrize(
    "pin",
    [
        b"9999",
        "9999",
    ],
)
def test_mastercard_pin_change(pin: Union[bytes, str]) -> None:
    """
    Test generation of offline PIN script with ISO format 2 PIN block

    Master Key Derivation = Option A
    SMI/SMC using Common Session Key Derivation (EMV SU-46)
    PIN encipherment with optional padding when block is not
    multiple of 8. However, no padding takes place because PIN
    block is 8 bytes long.
    """
    # Verify issuer master key check digits
    iss_mk_smc = bytes.fromhex("11111111111111112222222222222222")
    assert tools.key_check_digits(iss_mk_smc, 3).hex().upper() == "D2B91C"
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 3).hex().upper() == "08D7B4"

    # Derive ICC master keys
    pan = "1234567890123456"
    psn = "01"
    icc_mk_smc = kd.derive_icc_mk_a(iss_mk_smc, pan, psn)
    assert tools.key_check_digits(icc_mk_smc, 3).hex().upper() == "6B0F76"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 3).hex().upper() == "02396B"

    # Verify SMC/SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("ABCDEF1234567890")
    sk_smc = kd.derive_common_sk(icc_mk_smc, r)
    assert tools.key_check_digits(sk_smc, 3).hex().upper() == "D694B8"
    sk_smi = kd.derive_common_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 3).hex().upper() == "F088F8"

    # Encrypt new PIN
    pin_block = sm.format_iso9564_2_pin_block(pin)
    assert pin_block.hex().upper() == "249999FFFFFFFFFF"
    command_data = sm.encrypt_command_data(
        sk_smc, pin_block, sm.EncryptionType.MASTERCARD
    )
    assert command_data.hex().upper() == "28367E05DE3381C2"

    # Script MAC generated using Session Key
    command_header = bytes.fromhex("8424000210")
    mac = sm.generate_command_mac(sk_smi, command_header + command_data)
    assert mac.hex().upper() == "8C4D10091A093B5B"


@pytest.mark.parametrize(
    "pin",
    [
        b"999999999999",
        "999999999999",
    ],
)
def test_mastercard_pin_change_12(pin: Union[bytes, str]) -> None:
    """
    Test generation of offline PIN script with ISO format 2 PIN block
    PIN is 12 digits long.

    Master Key Derivation = Option A
    SMI/SMC using Common Session Key Derivation (EMV SU-46)
    PIN encipherment with optional padding when block is not
    multiple of 8. However, no padding takes place because PIN
    block is 8 bytes long.
    """
    # Verify issuer master key check digits
    iss_mk_smc = bytes.fromhex("11111111111111112222222222222222")
    assert tools.key_check_digits(iss_mk_smc, 3).hex().upper() == "D2B91C"
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 3).hex().upper() == "08D7B4"

    # Derive ICC master keys
    pan = "1234567890123456"
    psn = "01"
    icc_mk_smc = kd.derive_icc_mk_a(iss_mk_smc, pan, psn)
    assert tools.key_check_digits(icc_mk_smc, 3).hex().upper() == "6B0F76"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 3).hex().upper() == "02396B"

    # Verify SMC/SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("ABCDEF1234567890")
    sk_smc = kd.derive_common_sk(icc_mk_smc, r)
    assert tools.key_check_digits(sk_smc, 3).hex().upper() == "D694B8"
    sk_smi = kd.derive_common_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 3).hex().upper() == "F088F8"

    # Encrypt new PIN
    pin_block = sm.format_iso9564_2_pin_block(pin)
    assert pin_block.hex().upper() == "2C999999999999FF"
    command_data = sm.encrypt_command_data(
        sk_smc, pin_block, sm.EncryptionType.MASTERCARD
    )
    assert command_data.hex().upper() == "99AD3AF9EB3ECF0C"

    # Script MAC generated using Session Key
    command_header = bytes.fromhex("8424000210")
    mac = sm.generate_command_mac(sk_smi, command_header + command_data)
    assert mac.hex().upper() == "1471D88C6393BE21"


@pytest.mark.parametrize(
    "pin",
    [
        b"9999",
        "9999",
    ],
)
def test_vis_pin_change(pin: Union[bytes, str]) -> None:
    """
    Test generation of offline PIN script with VIS PIN block

    Master Key Derivation = Option A
    SMI/SMC using Visa Session Key Derivation
    PIN encipherment with mandatory Visa padding
    """
    # Verify issuer master key check digits
    iss_mk_ac = bytes.fromhex("FEDCBA98765432100123456789ABCDEF")
    assert tools.key_check_digits(iss_mk_ac, 3).hex().upper() == "7B8358"
    iss_mk_smc = bytes.fromhex("11111111111111112222222222222222")
    assert tools.key_check_digits(iss_mk_smc, 3).hex().upper() == "D2B91C"
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 3).hex().upper() == "08D7B4"

    # Derive ICC master keys
    pan = "1234567890123456"
    psn = "01"
    icc_mk_ac = kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
    assert tools.key_check_digits(icc_mk_ac, 3).hex().upper() == "C5CACD"
    icc_mk_smc = kd.derive_icc_mk_a(iss_mk_smc, pan, psn)
    assert tools.key_check_digits(icc_mk_smc, 3).hex().upper() == "6B0F76"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 3).hex().upper() == "02396B"

    # Verify SMC/SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("FFFF")
    sk_smc = kd.derive_visa_sm_sk(icc_mk_smc, r)
    assert tools.key_check_digits(sk_smc, 3).hex().upper() == "BD7C46"
    sk_smi = kd.derive_visa_sm_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 3).hex().upper() == "A41D47"

    # Encrypt new PIN
    pin_block = sm.format_vis_pin_block(icc_mk_ac, pin)
    assert pin_block.hex().upper() == "049999FFC4B001F1"
    command_data = sm.encrypt_command_data(sk_smc, pin_block, sm.EncryptionType.VISA)
    assert command_data.hex().upper() == "D421231A6FD2F0FAEE671384F0D3A7B9"

    # Script MAC generated using Session Key
    command_header = bytes.fromhex("8424000210")
    mac = sm.generate_command_mac(sk_smi, command_header + command_data)
    assert mac.hex().upper() == "F67E37E67B06AB99"


@pytest.mark.parametrize(
    ["pin", "current_pin"],
    [
        (b"9999", b"8888"),
        (b"9999", "8888"),
        ("9999", b"8888"),
        ("9999", "8888"),
    ],
)
def test_vis_change_change_current_pin(
    pin: Union[bytes, str], current_pin: Union[bytes, str]
) -> None:
    """
    Test generation of offline PIN script with VIS PIN block with current PIN

    Master Key Derivation = Option A
    SMI/SMC using Visa Session Key Derivation
    PIN encipherment with mandatory Visa padding
    """
    # Verify issuer master key check digits
    iss_mk_ac = bytes.fromhex("FEDCBA98765432100123456789ABCDEF")
    assert tools.key_check_digits(iss_mk_ac, 3).hex().upper() == "7B8358"
    iss_mk_smc = bytes.fromhex("11111111111111112222222222222222")
    assert tools.key_check_digits(iss_mk_smc, 3).hex().upper() == "D2B91C"
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 3).hex().upper() == "08D7B4"

    # Derive ICC master keys
    pan = "1234567890123456"
    psn = "01"
    icc_mk_ac = kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
    assert tools.key_check_digits(icc_mk_ac, 3).hex().upper() == "C5CACD"
    icc_mk_smc = kd.derive_icc_mk_a(iss_mk_smc, pan, psn)
    assert tools.key_check_digits(icc_mk_smc, 3).hex().upper() == "6B0F76"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 3).hex().upper() == "02396B"

    # Verify SMC/SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("FFFF")
    sk_smc = kd.derive_visa_sm_sk(icc_mk_smc, r)
    assert tools.key_check_digits(sk_smc, 3).hex().upper() == "BD7C46"
    sk_smi = kd.derive_visa_sm_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 3).hex().upper() == "A41D47"

    # Encrypt new PIN
    pin_block = sm.format_vis_pin_block(icc_mk_ac, pin, current_pin)
    assert pin_block.hex().upper() == "8C1199FFC4B001F1"
    command_data = sm.encrypt_command_data(sk_smc, pin_block, sm.EncryptionType.VISA)
    assert command_data.hex().upper() == "C29371CC36BA70C1EE671384F0D3A7B9"

    # Script MAC generated using Session Key
    command_header = bytes.fromhex("8424000110")
    mac = sm.generate_command_mac(sk_smi, command_header + command_data)
    assert mac.hex().upper() == "6AB3AD15D8B0621A"


@pytest.mark.parametrize(
    ["pin", "current_pin"],
    [
        (b"999999999999", b"888888888888"),
        (b"999999999999", "888888888888"),
        ("999999999999", b"888888888888"),
        ("999999999999", "888888888888"),
    ],
)
def test_vis_change_current_pin_12(
    pin: Union[bytes, str], current_pin: Union[bytes, str]
) -> None:
    """
    Test generation of offline PIN script with VIS PIN block with current PIN
    Both PINs are 12 digits long.

    Master Key Derivation = Option A
    SMI/SMC using Visa Session Key Derivation
    PIN encipherment with mandatory Visa padding
    """
    # Verify issuer master key check digits
    iss_mk_ac = bytes.fromhex("FEDCBA98765432100123456789ABCDEF")
    assert tools.key_check_digits(iss_mk_ac, 3).hex().upper() == "7B8358"
    iss_mk_smc = bytes.fromhex("11111111111111112222222222222222")
    assert tools.key_check_digits(iss_mk_smc, 3).hex().upper() == "D2B91C"
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 3).hex().upper() == "08D7B4"

    # Derive ICC master keys
    pan = "1234567890123456"
    psn = "01"
    icc_mk_ac = kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
    assert tools.key_check_digits(icc_mk_ac, 3).hex().upper() == "C5CACD"
    icc_mk_smc = kd.derive_icc_mk_a(iss_mk_smc, pan, psn)
    assert tools.key_check_digits(icc_mk_smc, 3).hex().upper() == "6B0F76"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 3).hex().upper() == "02396B"

    # Verify SMC/SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("FFFF")
    sk_smc = kd.derive_visa_sm_sk(icc_mk_smc, r)
    assert tools.key_check_digits(sk_smc, 3).hex().upper() == "BD7C46"
    sk_smi = kd.derive_visa_sm_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 3).hex().upper() == "A41D47"

    # Encrypt new PIN
    pin_block = sm.format_vis_pin_block(icc_mk_ac, pin, current_pin)
    assert pin_block.hex().upper() == "841111112A5E67F1"
    command_data = sm.encrypt_command_data(sk_smc, pin_block, sm.EncryptionType.VISA)
    assert command_data.hex().upper() == "622066A74F1E974AEE671384F0D3A7B9"

    # Script MAC generated using Session Key
    command_header = bytes.fromhex("8424000110")
    mac = sm.generate_command_mac(sk_smi, command_header + command_data)
    assert mac.hex().upper() == "829AC5DE4B88074D"


@pytest.mark.parametrize(
    "pin",
    [
        b"9999",
        "9999",
    ],
)
def test_ccd_pin_change(pin: Union[bytes, str]) -> None:
    """
    Test generation of offline PIN script with ISO format 2 PIN block'
    Common Core Definition. Mandatory padding.

    Master Key Derivation = Option A
    SMI/SMC using Common Session Key Derivation (EMV SU-46)
    PIN encipherment with optional padding when block is not
    multiple of 8. However, no padding takes place because PIN
    block is 8 bytes long.
    """
    # Verify issuer master key check digits
    iss_mk_smc = bytes.fromhex("11111111111111112222222222222222")
    assert tools.key_check_digits(iss_mk_smc, 3).hex().upper() == "D2B91C"
    iss_mk_smi = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk_smi, 3).hex().upper() == "08D7B4"

    # Derive ICC master keys
    pan = "1234567890123456"
    psn = "01"
    icc_mk_smc = kd.derive_icc_mk_a(iss_mk_smc, pan, psn)
    assert tools.key_check_digits(icc_mk_smc, 3).hex().upper() == "6B0F76"
    icc_mk_smi = kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
    assert tools.key_check_digits(icc_mk_smi, 3).hex().upper() == "02396B"

    # Verify SMC/SMI session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("ABCDEF1234567890")
    sk_smc = kd.derive_common_sk(icc_mk_smc, r)
    assert tools.key_check_digits(sk_smc, 3).hex().upper() == "D694B8"
    sk_smi = kd.derive_common_sk(icc_mk_smi, r)
    assert tools.key_check_digits(sk_smi, 3).hex().upper() == "F088F8"

    # Encrypt new PIN
    pin_block = sm.format_iso9564_2_pin_block(pin)
    assert pin_block.hex().upper() == "249999FFFFFFFFFF"
    command_data = sm.encrypt_command_data(sk_smc, pin_block, sm.EncryptionType.EMV)
    assert command_data.hex().upper() == "28367E05DE3381C29D30A6055A8DEB86"

    # Script MAC generated using Session Key
    command_header = bytes.fromhex("8424000210")
    mac = sm.generate_command_mac(sk_smi, command_header + command_data)
    assert mac.hex().upper() == "69DA1EE208823022"



r"""Provide functions to generate Dynamic Card Verification Values.
Dynamic Card Verification is used in EMV transactions to
generate a new CVV for each transaction.
"""


from pyemv import mac as _mac
from pyemv import tools as _tools

__all__ = ["generate_cvc3"]


def generate_cvc3(icc_cvc3: bytes, track_template: bytes, atc: bytes, un: bytes) -> str:
    r"""Generate MasterCard Dynamic Card Verification Code (CVC3).

    Parameters
    ----------
    icc_cvc3 : bytes
        Binary 16-byte ICC CVC3 key encoded on the card.
        This key is derived from Issuer Master CVC3 key using
        either `kd.derive_icc_mk_a` or `kd.derive_icc_mk_b`.
        Has to be a valid Triple DES key.
    track_template : bytes
        Binary track template data.
        Track2 template consists of:

            PAN || "D" || EXPIRY DATE || SERVICE CODE || DD || "F"?

        DD is Discretionary Data up to 13 digits. DD must have
        value before it's encoded with ATC, UN and CVC3.
        If the resulting track2 template is odd length then it must
        be padded with an "F".

        Track1 template consists of:

            "B" || PAN || ^LAST/FIRST^ || EXPIRY DATE || SERVICE CODE || DD

        Track1 template must be encoded as an ASCII byte string, where, for example,
        character "B" is encoded as "\x42".

    atc : bytes
        Binary data from tag 9F36 (Application Transaction Counter).
    un : bytes
        Binary data from tag 9F37 (Unpredictable Number).

    Returns
    -------
    cvc3 : str
        5-digit Dynamic Card Verification Code.
        Compare generated CVC3 against number of available
        least significant digits of CVC3 received on track1/2.

    Raises
    ------
    ValueError
        ICC CVC3 key must be a double length DES key
    ValueError
        ATC value must be 2 bytes long
    ValueError
        Unpredictable number must be 4 bytes long

    Examples
    --------
    >>> from pyemv.cvv import generate_cvc3
    >>> from pyemv.kd import derive_icc_mk_a
    >>> iss_cvc3 = bytes.fromhex("01234567899876543210012345678998")
    >>> pan = "5123456789012345"
    >>> psn = "00"
    >>> icc_cvc3 = derive_icc_mk_a(iss_cvc3, pan, psn)
    >>> track2 = bytes.fromhex("5123456789012345D35121010000000000000F")
    >>> atc = bytes.fromhex("005E")
    >>> un = bytes.fromhex("00000899")
    >>> generate_cvc3(icc_cvc3, track2, atc, un)
    '29488'
    """
    if len(icc_cvc3) != 16:
        raise ValueError("ICC CVC3 key must be a double length DES key")

    if len(atc) != 2:
        raise ValueError("ATC value must be 2 bytes long")

    if len(un) != 4:
        raise ValueError("Unpredictable number must be 4 bytes long")

    # IV CVC3 is formed from 2 least significant bytes of CBC-MAC
    # computed over track template
    iv_cvc3 = _mac.mac_iso9797_3(icc_cvc3[:8], icc_cvc3[-8:], track_template, 2)[-2:]

    # CVC3 is formed from 2 least significant bytes of TDES encrypted
    # ciphertext block
    block = iv_cvc3 + un + atc
    cvc3 = _tools.encrypt_tdes_ecb(icc_cvc3, block)[-2:]

    return str(int(cvc3.hex(), 16))



r"""PyEMV package provides methods to generate

    - Application Cryptograms (TC, ARQC, or AAC) used to verify ICC
    - Authorization Response Cryptogram (ARPC) used to verify card issuer
    - Secure Messaging Integrity and Confidentiality used by the issuer to update values on the ICC
    - Dynamic Card Verification Values

PyEMV consists of the following modules:

    - kd - Key Derivation support for ICC master keys and session keys.
    - ac - Application Cryptogram support for ARQC, AAC, TC, and ARPC.
    - sm - Secure Messaging support for script command integrity and confidentiality.
    - cvn - Putting it all together for various Cryptogram Version Numbers.
    - cvv - Support for dynamic card verification, such as CVC3.
    - tlv - TLV encoder and decoder.

For example, ICC Master Key derivation method A:

    >>> import pyemv
    >>> icc_mk = pyemv.kd.derive_icc_mk_a(
    ...     iss_mk=bytes.fromhex('0123456789ABCDEFFEDCBA9876543210'),
    ...     pan='99012345678901234',
    ...     psn='45')
    >>> icc_mk.hex().upper()
    '67F8292358083E5EA7AB7FDA58D53B6B'
"""

__version__ = "1.5.0"
__author__ = "Konstantin Novichikhin <konstantin.novichikhin@gmail.com>"

from pyemv import ac, cvn, cvv, kd, mac, sm, tlv, tools



import pytest

from pyemv import mac


def test_mac_iso9797_3() -> None:
    # SK < 16 bytes
    with pytest.raises(
        ValueError,
        match="pecify valid padding method: 1 or 2.",
    ):
        mac.mac_iso9797_3(
            bytes.fromhex("AAAAAAAAAAAAAAAA"),
            bytes.fromhex("BBBBBBBBBBBBBBBB"),
            b"hello world",
            4,
        )

    _mac = (
        mac.mac_iso9797_3(
            bytes.fromhex("AAAAAAAAAAAAAAAA"),
            bytes.fromhex("BBBBBBBBBBBBBBBB"),
            b"hello world",
            1,
        )
        .hex()
        .upper()
    )
    assert _mac == "78DA3D3DEB48FD4D"

    _mac = (
        mac.mac_iso9797_3(
            bytes.fromhex("AAAAAAAAAAAAAAAA"),
            bytes.fromhex("BBBBBBBBBBBBBBBB"),
            b"hello world",
            2,
        )
        .hex()
        .upper()
    )
    assert _mac == "2BC2D9EDE0CF31F6"


def test_pad_iso9797_1() -> None:
    _mac = mac.pad_iso9797_1(b"")
    assert _mac == b"\x00" * 8

    for i in range(0, 9):
        _mac = mac.pad_iso9797_1(b"F" * i, 8)
        assert _mac == (b"F" * i) + (b"\x00" * (8 - i))

    for i in range(0, 17):
        _mac = mac.pad_iso9797_1(b"F" * i, 16)
        assert _mac == (b"F" * i) + (b"\x00" * (16 - i))


def test_pad_iso9797_2() -> None:
    _mac = mac.pad_iso9797_2(b"")
    assert _mac == b"\x80" + b"\x00" * 7

    for i in range(0, 8):
        _mac = mac.pad_iso9797_2(b"F" * i, 8)
        assert _mac == (b"F" * i) + b"\x80" + (b"\x00" * (7 - i))
    _mac = mac.pad_iso9797_2(b"F" * 8, 8)
    assert _mac == (b"F" * 8) + b"\x80" + (b"\x00" * 7)

    for i in range(0, 16):
        _mac = mac.pad_iso9797_2(b"F" * i, 16)
        assert _mac == (b"F" * i) + b"\x80" + (b"\x00" * (15 - i))
    _mac = mac.pad_iso9797_2(b"F" * 16, 16)
    assert _mac == (b"F" * 16) + b"\x80" + (b"\x00" * 15)



import sys as _sys
import typing as _typing

from cryptography.hazmat.backends import default_backend as _default_backend
from cryptography.hazmat.primitives.ciphers import Cipher as _Cipher
from cryptography.hazmat.primitives.ciphers import algorithms as _algorithms
from cryptography.hazmat.primitives.ciphers import modes as _modes

__all__ = [
    "xor",
    "odd_parity",
    "adjust_key_parity",
    "key_check_digits",
    "encrypt_tdes_cbc",
    "encrypt_tdes_ecb",
]


def xor(data: bytes, key: bytes) -> bytes:
    r"""Apply "exlusive or" to two bytes instances.

    Many thanks:
    https://stackoverflow.com/a/29409299

    Parameters
    ----------
    data : bytes
        Data to be XOR'd
    key : bytes
        Bit mask used to XOR data

    Returns
    -------
    bytes
        Data XOR'd by key
    """
    key = key[: len(data)]
    int_var = int.from_bytes(data, _sys.byteorder)
    int_key = int.from_bytes(key, _sys.byteorder)
    int_enc = int_var ^ int_key
    return int_enc.to_bytes(len(data), _sys.byteorder)


def odd_parity(v: int) -> int:
    r"""Check integer parity.

    Many thanks: in_parallel
    http://p-nand-q.com/python/_algorithms/math/bit-parity.html

    Parameters
    ----------
    v : int
        Integer to check parity of

    Returns
    -------
    int
        0 = even parity (even number of bits enabled, e.g. 0, 3, 5)
        1 = odd parity (odd number of bits enabled, e.g. 1, 2, 4)
    """
    v ^= v >> 16
    v ^= v >> 8
    v ^= v >> 4
    v &= 0xF
    return (0x6996 >> v) & 1


def adjust_key_parity(key: _typing.Union[bytes, bytearray]) -> bytes:
    r"""Adjust DES key parity key.

    Parameters
    ----------
    key : bytes, bytearray
        Binary key to adjust for odd parity.

    Returns
    -------
    adjusted_key : bytes
        Binary key adjusted for odd parity.

    Examples
    --------
    >>> from pyemv import tools
    >>> key = bytes.fromhex("1A2B3C4D5F0A1B2C4D5F6A7B8C9D0F1A")
    >>> tools.adjust_key_parity(key).hex().upper()
    '1A2A3D4C5E0B1A2C4C5E6B7A8C9D0E1A'
    """
    adjusted_key = bytearray(key)

    for i, byte in enumerate(adjusted_key):
        if not odd_parity(byte):
            adjusted_key[i] ^= 1

    return bytes(adjusted_key)


def key_check_digits(key: bytes, length: int = 2) -> bytes:
    r"""Calculate Triple DES key check digits.

    Parameters
    ----------
    key : bytes
        Binary key to provide check digits for. Has to be a valid DES key.
    length : int, optional
        Number of key check digits bytes provided in the response (default 2).

    Returns
    -------
    check_digits : bytes
        Binary check digits (`length` bytes)

    Examples
    --------
    >>> from pyemv import tools
    >>> key = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> tools.key_check_digits(key).hex().upper()
    '08D7'
    """
    cipher = _Cipher(
        _algorithms.TripleDES(key), _modes.ECB(), backend=_default_backend()
    )
    encryptor = cipher.encryptor()
    return encryptor.update(b"\x00\x00\x00\x00\x00\x00\x00\x00")[:length]


def encrypt_tdes_cbc(key: bytes, iv: bytes, data: bytes) -> bytes:
    r"""Encrypt data using Triple DES CBC algorithm.

    Parameters
    ----------
    key : bytes
        Binary Triple DES key. Has to be a valid DES key.
    iv : bytes
        Binary initial initialization vector for CBC.
    data : bytes
        Binary data to be encrypted.

    Returns
    -------
    encrypted_data : bytes
        Binary encrypted data.

    Examples
    --------
    >>> from pyemv.tools import encrypt_tdes_cbc
    >>> key = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> iv = bytes.fromhex("0000000000000000")
    >>> encrypt_tdes_cbc(key, iv, b"12345678").hex().upper()
    '41D2FFBA3CDC15FE'
    """
    cipher = _Cipher(
        _algorithms.TripleDES(key),
        _modes.CBC(iv),
        backend=_default_backend(),
    )
    return cipher.encryptor().update(data)


def encrypt_tdes_ecb(key: bytes, data: bytes) -> bytes:
    r"""Encrypt data using Triple DES ECB algorithm.

    Parameters
    ----------
    key : bytes
        Binary Triple DES key. Has to be a valid DES key.
    data : bytes
        Binary data to be encrypted.

    Returns
    -------
    encrypted_data : bytes
        Binary encrypted data.

    Examples
    --------
    >>> from pyemv.tools import encrypt_tdes_ecb
    >>> key = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> encrypt_tdes_ecb(key, b"12345678").hex().upper()
    '41D2FFBA3CDC15FE'
    """
    cipher = _Cipher(
        _algorithms.TripleDES(key), _modes.ECB(), backend=_default_backend()
    )
    return cipher.encryptor().update(data)



import typing as _typing

from cryptography.hazmat.backends import default_backend as _default_backend
from cryptography.hazmat.primitives.ciphers import Cipher as _Cipher
from cryptography.hazmat.primitives.ciphers import algorithms as _algorithms
from cryptography.hazmat.primitives.ciphers import modes as _modes

__all__ = [
    "mac_iso9797_3",
    "pad_iso9797_1",
    "pad_iso9797_2",
]


def mac_iso9797_3(
    key1: bytes,
    key2: bytes,
    data: bytes,
    padding: int,
    length: _typing.Optional[int] = None,
) -> bytes:
    r"""ISO/IEC 9797-1 MAC algorithm 3. Requires two independent keys.
    Only the last data block is processed using TDES,
    all previous blocks are processed using single DES.

    Parameters
    ----------
    key1 : bytes
        Binary MAC key used in initial transformation.
        Has to be a valid DES key.
    key2 : bytes
        Binary MAC key used  in output transformation.
        Has to be a valid DES key.
    data : bytes
        Data to be MAC'd.
    padding : int
        Padding method of `data`.

            - 1 = ISO/IEC 9797-1 method 1.
            - 2 = ISO/IEC 9797-1 method 2.

    length : int, optional
        Desired length of AC [4 <= N <= 8] (default 8 bytes).

    Returns
    -------
    mac : bytes
        Returns a binary MAC of requested length

    Raises
    ------
    ValueError
        Invalid padding method specified

    Notes
    -----
    See https://en.wikipedia.org/wiki/ISO/IEC_9797-1 for the
    algorithm reference.

    See Also
    --------
    pyemv.mac.pad_iso9797_1 : ISO/IEC 9791-1 padding method 1
    pyemv.mac.pad_iso9797_2 : ISO/IEC 9791-1 padding method 2
    pyemv.mac.pad_iso9797_3 : ISO/IEC 9791-1 padding method 3

    Examples
    --------
    >>> from pyemv.mac import mac_iso9797_3
    >>> key1 = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> key2 = bytes.fromhex("FEDCBA98765432100123456789ABCDEF")
    >>> data = bytes.fromhex("1234567890ABCDEF")
    >>> mac_iso9797_3(key1, key2, data, padding=2).hex().upper()
    '644AA5C915DBDAF8'
    """
    if length is None:
        length = 8

    if padding == 1:
        data = pad_iso9797_1(data, 8)
    elif padding == 2:
        data = pad_iso9797_2(data, 8)
    else:
        raise ValueError("Specify valid padding method: 1 or 2.")

    # Encrypt first block with key1 then
    # encrypt the rest of the data in CBC mode
    cipher1 = _Cipher(
        _algorithms.TripleDES(key1),
        _modes.CBC(b"\x00\x00\x00\x00\x00\x00\x00\x00"),
        backend=_default_backend(),
    )
    encryptor1 = cipher1.encryptor()
    data = encryptor1.update(data)[-8:]

    # Decrypt the last block with key2 and then encrypt it with key1
    cipher2 = _Cipher(
        _algorithms.TripleDES(key2), _modes.CBC(data), backend=_default_backend()
    )
    decryptor2 = cipher2.decryptor()
    return encryptor1.update(decryptor2.update(data))[:length]


def pad_iso9797_1(data: bytes, block_size: _typing.Optional[int] = None) -> bytes:
    r"""ISO/IEC 9797-1 padding method 1.
    Add the smallest number of "0x00" bytes to the right
    such that the length of resulting message is a multiple of
    `block_size` bytes. If the data is already multiple of
    `block_size` bytes then no bytes added

    Parameters
    ----------
    data : bytes
        Data to be padded
    block_size : int, optional
        Padded data will be multiple of specified block size (default 8).

    Returns
    -------
    bytes
        Padded data

    Notes
    -----
    See https://en.wikipedia.org/wiki/ISO/IEC_9797-1 for the
    algorithm reference.

    Examples
    --------
    >>> from pyemv.mac import pad_iso9797_1
    >>> pad_iso9797_1(bytes.fromhex("1234")).hex().upper()
    '1234000000000000'
    """
    if block_size is None:
        block_size = 8

    remainder = len(data) % block_size
    if remainder > 0:
        return data + (b"\x00" * (block_size - remainder))

    if len(data) == 0:
        return b"\x00" * block_size

    return data


def pad_iso9797_2(data: bytes, block_size: _typing.Optional[int] = None) -> bytes:
    r"""ISO/IEC 9797-1 padding method 2 (equivalent to ISO/IEC 7816-4).
    Add a mandatory "0x80" byte to the right of data,
    and then add the smallest number of "0x00" bytes to the right
    such that the length of resulting message is a multiple of
    `block_size` bytes.

    Parameters
    ----------
    data : bytes
        Data to be padded
    block_size : int, optional
        Padded data will be multiple of specified block size (default 8).

    Returns
    -------
    bytes
        Padded data

    Notes
    -----
    See https://en.wikipedia.org/wiki/ISO/IEC_9797-1 for the
    algorithm reference.

    Examples
    --------
    >>> from pyemv.mac import pad_iso9797_2
    >>> pad_iso9797_2(bytes.fromhex("1234")).hex().upper()
    '1234800000000000'
    """
    if block_size is None:
        block_size = 8

    return pad_iso9797_1(data + b"\x80", block_size)



r"""Use TLV decoder and encoder to disassemble and assemble tag-length-value EMV data.

By default TLV data is broken down into bytes:

    >>> import pyemv
    >>> tlv_data = bytes.fromhex("9C0101E0055F2A0202089F0200")
    >>> pyemv.tlv.decode(tlv_data)
    {'9C': b'\x01', 'E0': {'5F2A': b'\x02\x08'}, '9F02': b''}
    >>> pyemv.tlv.encode({'9C': b'\x01', 'E0': {'5F2A': b'\x02\x08'}, '9F02': b''}).hex().upper()
    '9C0101E0055F2A0202089F0200'

It can also be converted to strings (among other things):

    >>> import pyemv
    >>> tlv_data = bytes.fromhex("9C0101E0055F2A0202089F0200")
    >>> pyemv.tlv.decode(tlv_data, convert=lambda t, v: v.hex().upper())
    {'9C': '01', 'E0': {'5F2A': '0208'}, '9F02': ''}
    >>> pyemv.tlv.encode({'9C': '01', 'E0': {'5F2A': '0208'}, '9F02': ''}).hex().upper()
    '9C0101E0055F2A0202089F0200'
"""

import typing as _t

__all__ = ["decode", "DecodeError", "encode", "EncodeError"]


class DecodeError(ValueError):
    r"""Subclass of ValueError that describes TLV decoding error.

    Attributes
    ----------
    msg : str
        The unformatted error message
    tag : str
        Tag where decoding stopped
    offset : int
        Offset in the input data where decoding stopped
    tlv : dict
        Dictionary with partially decoded data
    """

    def __init__(
        self,
        msg: str,
        tag: str,
        offset: int,
        tlv: _t.Dict[str, _t.Any],
    ):
        errmsg = f"{msg}: tag '{tag}', offset {offset}."
        ValueError.__init__(self, errmsg)
        self.msg = msg
        self.tag = tag
        self.offset = offset
        self.tlv = tlv


class EncodeError(ValueError):
    r"""Subclass of ValueError that describes TLV encoding error.

    Attributes
    ----------
    msg : str
        The unformatted error message
    tag : str
        Tag where decoding stopped
    """

    def __init__(
        self,
        msg: str,
        tag: str,
    ):
        errmsg = f"{msg}: tag '{tag}'."
        ValueError.__init__(self, errmsg)
        self.msg = msg
        self.tag = tag


# fmt: off
_S = _t.TypeVar("_S")
@_t.overload
def decode(data: bytes) -> _t.Dict[str, _t.Any]: ...
@_t.overload
def decode(data: bytes, *, simple: _t.Optional[bool]) -> _t.Dict[str, _t.Any]: ...
@_t.overload
def decode(data: bytes, *, convert: _t.Optional[_t.Callable[[str, _t.Union[bytes, bytearray]], _t.Any]]) -> _t.Dict[str, _t.Any]: ...
@_t.overload
def decode(data: bytes, *, simple: _t.Optional[bool], convert: _t.Optional[_t.Callable[[str, _t.Union[bytes, bytearray]], _t.Any]]) -> _t.Dict[str, _t.Any]: ...
@_t.overload
def decode(data: bytes, *, flatten: _t.Optional[bool] = True) -> _t.Dict[str, bytes]: ...
@_t.overload
def decode(data: bytes, *, flatten: _t.Optional[bool] = True, convert: _t.Optional[_t.Callable[[str, _t.Union[bytes, bytearray]], _S]]) -> _t.Dict[str, _S]: ...
@_t.overload
def decode(data: bytes, *, flatten: _t.Optional[bool] = True, simple: _t.Optional[bool]) -> _t.Dict[str, bytes]: ...
@_t.overload
def decode(data: bytes, *, flatten: _t.Optional[bool] = True, simple: _t.Optional[bool], convert: _t.Optional[_t.Callable[[str, _t.Union[bytes, bytearray]], _S]]) -> _t.Dict[str, _S]: ...
# fmt: on


def decode(
    data: _t.Union[bytes, bytearray],
    *,
    flatten: _t.Optional[bool] = None,
    simple: _t.Optional[bool] = None,
    convert: _t.Optional[_t.Callable[[str, _t.Union[bytes, bytearray]], _t.Any]] = None,
) -> _t.Dict[str, _t.Any]:
    r"""Decode TLV data.

    Parameters
    ----------
    data : bytes
        Encoded TLV data
    flatten : bool, optional
        Flatten constructed tags and return one flat dictionary
        with all tags together. Defaults to False.
    simple : bool, optional
        Some specification stipulate that TLV length is always
        1 byte long with a maximum length of 255.
        To enable this option set simple to True. Defaults to False.
    convert : callable, optional
        Apply this function to every primitive tag value and
        return tag value in desired format.
        Function must accept tag name as a first argument and
        tag value as a second argument.
        Defauls to 'lambda t, v: bytes(v)' to return bytes objects.

    Returns
    -------
    tlv : dict
        Dictionary with decoded data

    Raises
    ------
    DecodeError

    Notes
    -----
    This decoder adheres to Rules for BER-TLV Data Objects in Annex B or
    EMV 4.3 Book 3 Application Specification.

    Examples
    --------
    >>> from pyemv import tlv
    >>> tlv.decode(bytes.fromhex("9C0101E0055F2A0202089F0200"))
    {'9C': b'\x01', 'E0': {'5F2A': b'\x02\x08'}, '9F02': b''}
    >>> tlv.decode(bytes.fromhex("9C0101E0055F2A0202089F0200"), flatten=True)
    {'9C': b'\x01', '5F2A': b'\x02\x08', '9F02': b''}
    >>> tlv.decode(bytes.fromhex("9C0101E0055F2A0202089F0200"), convert=lambda t, v: v.hex().upper())
    {'9C': '01', 'E0': {'5F2A': '0208'}, '9F02': ''}
    >>> tlv.decode(bytes.fromhex("9C0101E0055F2A0202089F0200"), flatten=True, convert=lambda t, v: v.hex().upper())
    {'9C': '01', '5F2A': '0208', '9F02': ''}
    """

    if flatten is None:
        flatten = False

    if simple is None:
        simple = False

    if convert is None:
        convert = lambda t, v: bytes(v)

    dec: _t.Dict[str, _t.Any] = {}

    try:
        _decode(data, 0, len(data), dec, flatten, simple, convert)
    except DecodeError as e:
        # Catch the error here to provide reference
        # to a partically decoded data.
        e.tlv = dec
        raise
    return dec


def _decode(
    data: _t.Union[bytes, bytearray],
    ofst: int,
    ofst_limit: int,
    dec: _t.Dict[str, _t.Any],
    flatten: bool,
    simple: bool,
    convert: _t.Callable[[str, _t.Union[bytes, bytearray]], _S],
) -> int:
    while ofst < ofst_limit:
        # Determine tag name length.
        tag_name_len = 1
        try:
            # If b0-4 are on then a 2nd byte follows.
            constructed = bool(data[ofst] & 0b00100000)
            if (data[ofst] & 0b00011111) == 0b00011111:
                # If b7 is on then another byte follows
                while data[ofst + tag_name_len] & 0b10000000:
                    tag_name_len += 1
                tag_name_len += 1
        except IndexError:
            raise DecodeError(
                "Tag malformed, expecting more data",
                data[ofst : ofst + tag_name_len].hex().upper(),
                ofst,
                dec,
            ) from None

        # Check that tag name falls within parent tag
        if ofst + tag_name_len > ofst_limit:
            raise DecodeError(
                "Tag malformed, expecting more data",
                data[ofst : min(ofst + tag_name_len, ofst_limit)].hex().upper(),
                ofst,
                dec,
            )

        # Save tag name and move farther
        tag = data[ofst : ofst + tag_name_len].hex().upper()
        ofst += tag_name_len

        # Determine tag length
        tag_len_len = 1

        # Check that tag length falls within parent tag
        if ofst + tag_len_len > ofst_limit:
            raise DecodeError(
                f"Tag length malformed, expecting {str(tag_len_len)} byte(s)",
                tag,
                ofst,
                dec,
            )

        if data[ofst] & 0b10000000 and not simple:
            tag_len_len = data[ofst] & 0b01111111
            ofst += 1
            # Data does not have enough bytes to contain full
            # length as indicated by the previous byte.
            if ofst + tag_len_len > ofst_limit:
                raise DecodeError(
                    f"Tag length malformed, expecting {str(tag_len_len)} byte(s)",
                    tag,
                    ofst,
                    dec,
                )
            tag_len = int.from_bytes(data[ofst : ofst + tag_len_len], "big")
            ofst += tag_len_len
        else:
            tag_len = data[ofst]
            ofst += tag_len_len

        # Check that tag data falls within parent tag
        if ofst + tag_len > ofst_limit:
            raise DecodeError(
                f"Tag value malformed, expecting {str(tag_len)} byte(s)",
                tag,
                ofst,
                dec,
            )

        # Constructed data type (b5=on)
        if constructed:
            if flatten:
                ofst = _decode(
                    data, ofst, ofst + tag_len, dec, flatten, simple, convert
                )
            else:
                dec[tag] = {}
                ofst = _decode(
                    data, ofst, ofst + tag_len, dec[tag], flatten, simple, convert
                )
        # Primitive data type
        else:
            dec[tag] = convert(tag, data[ofst : ofst + tag_len])
            ofst += tag_len

    return ofst


def encode(
    tlv: _t.Mapping[str, _t.Any],
    *,
    simple: _t.Optional[bool] = None,
) -> bytes:
    r"""Encode TLV data.

    Parameters
    ----------
    data : bytes
        Encoded TLV data
    simple : bool, optional
        Some specification stipulate that TLV length is always
        1 byte long with a maximum length of 255.
        To enable this option set simple to True. Defaults to False.

    Returns
    -------
    tlv : bytes
        Encoded TLV data

    Raises
    ------
    EncodeError

    Notes
    -----
    This encoder adheres to Rules for BER-TLV Data Objects in Annex B or
    EMV 4.3 Book 3 Application Specification.

    Examples
    --------
    >>> from pyemv import tlv
    >>> tlv_data = {'9C': b'\x01', 'E0': {'5F2A': b'\x02\x08'}, '9F02': b''}
    >>> tlv.encode(tlv_data).hex().upper()
    '9C0101E0055F2A0202089F0200'
    >>> tlv_data = {'9C': '01', 'E0': {'5F2A': '0208'}, '9F02': ''}
    >>> tlv.encode(tlv_data).hex().upper()
    '9C0101E0055F2A0202089F0200'
    """

    if simple is None:
        simple = False

    return bytes(_encode(tlv, simple))


def _encode(tlv: _t.Mapping[str, _t.Any], simple: bool) -> bytearray:
    data = bytearray()
    for tag_s, value in tlv.items():
        # Tag
        try:
            tag = bytes.fromhex(tag_s)
            data += tag
        except ValueError:
            raise EncodeError("Invalid tag format, expecting hexchar string", tag_s)

        # Check tag format
        try:
            # If b0-4 are on then a 2nd byte follows.
            tag_name_len = 1
            if (tag[0] & 0b00011111) == 0b00011111:
                # If b7 is on then another byte follows
                while tag[tag_name_len] & 0b10000000:
                    tag_name_len += 1
                tag_name_len += 1
        except IndexError:
            raise EncodeError(
                "Invalid tag format, expecting more data", tag_s
            ) from None

        if len(tag) != tag_name_len:
            raise EncodeError("Invalid tag format, extra data", tag_s)

        # Value
        # Constructed
        if bool(tag[0] & 0b00100000):
            if not isinstance(value, _t.Mapping):
                raise EncodeError(
                    f"Invalid value type ({value.__class__.__name__}) "
                    "for a constructed tag, expecting a dict",
                    tag_s,
                )
            value = _encode(value, simple)
        # Primitive
        elif isinstance(value, str):
            try:
                value = bytes.fromhex(value)
            except ValueError:
                raise EncodeError(
                    "Invalid value format, expecting hexchar string", tag_s
                )
        elif not isinstance(value, (bytes, bytearray)):
            raise EncodeError(
                f"Invalid value type ({value.__class__.__name__}) "
                "for a primitive tag, expecting bytes or str",
                tag_s,
            )

        # Length
        tag_len_len = 1

        if len(value) > 255 and simple:
            raise EncodeError(
                f"Value length ({str(len(value))}) "
                "cannot exceed 255 bytes when 'simple' is enabled",
                tag_s,
            )

        # Multi-byte length required
        if len(value) > 127 and not simple:
            while len(value) > 2 ** (8 * tag_len_len) - 1:
                tag_len_len += 1
            data += int.to_bytes(tag_len_len | 0b10000000, 1, "big")

        data += int.to_bytes(len(value), tag_len_len, "big") + value

    return data



r"""Provide functions for the generation of the Application
Cryptograms (TC, ARQC, or AAC) generated by the ICC and the
Authorisation Response Cryptogram (ARPC) generated by the
issuer and verified by the ICC.

ARQC:

    >>> import pyemv
    >>> session_key = bytes.fromhex('29B33180E567CE38EA4CBC9D753B0E61')
    >>> ac_data = bytes.fromhex('0123456789ABCDEF0123456789ABCDEF')
    >>> arqc = pyemv.ac.generate_ac(session_key, ac_data)
    >>> arqc.hex().upper()
    'FA624250B008B59A'

ARPC:

    >>> import pyemv
    >>> session_key = bytes.fromhex('29B33180E567CE38EA4CBC9D753B0E61')
    >>> arpc_rc = bytes.fromhex('0000')
    >>> arpc = pyemv.ac.generate_arpc_1(session_key, arqc, arpc_rc)
    >>> arpc.hex().upper()
    '45D4255EEF10C920'
    >>> csu = bytes.fromhex('00000000')
    >>> arpc = pyemv.ac.generate_arpc_2(session_key, arqc, csu)
    >>> arpc.hex().upper()
    'CB56FA40'
"""

import typing as _typing
from enum import Enum as _Enum

from pyemv.mac import mac_iso9797_3 as _mac_iso9797_3
from pyemv.tools import encrypt_tdes_cbc as _encrypt_tdes_cbc
from pyemv.tools import xor as _xor

__all__ = [
    "PaddingType",
    "generate_ac",
    "generate_arpc_1",
    "generate_arpc_2",
]


class PaddingType(_Enum):
    VISA = 1
    EMV = 2


def generate_ac(
    sk_ac: bytes,
    data: bytes,
    padding_type: _typing.Optional[PaddingType] = None,
    length: _typing.Optional[int] = None,
) -> bytes:
    r"""First and second AC Generation using 8-byte block chiper
    Triple DES with ISO/IEC 9797-1 Algorithm 3. Same process for

        - Authorisation Request Cryptogram (ARQC)
        - Transaction Cryptogram (TC)
        - Application Authentication Cryptogram (AAC)

    Parameters
    ----------
    sk_ac : bytes
        Binary ICC Session Key for AC. Has to be a valid DES key.
    data : bytes
        Binary data from tags used in AC generation. Minimum recommended
        tags by EMV Book2 Application Cryptogram Data Selection.
        +------+-----+----------+---------------------------------+
        | Tag  | Len | Source   | Data                            |
        +------+-----+----------+---------------------------------+
        | 9F02 |   6 | Terminal | Amount, Authorized              |
        +------+-----+----------+---------------------------------+
        | 9F03 |   6 | Terminal | Amount, Other                   |
        +------+-----+----------+---------------------------------+
        | 9F1A |   2 | Terminal | Terminal Country Code           |
        +------+-----+----------+---------------------------------+
        | 95   |   5 | Terminal | Terminal Verification Results   |
        +------+-----+----------+---------------------------------+
        | 5F2A |   2 | Terminal | Transaction Currency Code       |
        +------+-----+----------+---------------------------------+
        | 9A   |   3 | Terminal | Transaction Date                |
        +------+-----+----------+---------------------------------+
        | 9C   |   1 | Terminal | Transaction Type                |
        +------+-----+----------+---------------------------------+
        | 9F37 |   4 | Terminal | Unpredictable Number            |
        +------+-----+----------+---------------------------------+
        | 82   |   2 | ICC      | Application Interchange Profile |
        +------+-----+----------+---------------------------------+
        | 9F36 |   2 | ICC      | Application Transaction Counter |
        +------+-----+----------+---------------------------------+
    padding_type : PaddingType, optional
        Padding method applied to the `data` (default EMV).

            - VISA = ISO/IEC 9797-1 method 1.
              Generally used by Visa schemes; not all of them.
            - EMV = ISO/IEC 9797-1 method 2.
              Generally used by other EMV schemes.

    length : int, optional
        Desired length of AC [4 <= `length` <= 8] (default 8 bytes).

    Returns
    -------
    ac : bytes
        Returns an `length`-byte cryptogram (ARQC, TC, AAC).

    Raises
    ------
    ValueError
        Session Key must be a double length DES key
    TypeError
        Padding type must be PaddingType Enum

    Notes
    -----
    The cryptogram takes as input a number of terminal and
    card data elements which are encrypted using the session key.
    The same elements must be sent to the host in the authorisation
    request so that the host can validate the cryptogram (also sent).

    After the terminal/reader receives the host generated ARPC in the
    response message it will send it to the card in the 2nd Generate AC
    command. The card will validate the ARPC and calculate another
    cryptogram (TC or AAC) which is returned in the 2nd Generate AC
    command. The calculation of this cryptogram is the same as for the
    1st cryptogram, using the same data input (although some of the
    values will be different, i.e. the CVR values).

    For further details see also:
        - EMV 4.3 Book 2 Section 8.1 Application Cryptogram Generation
        - EMV 4.3 Book 2 Annex A 1.2

    Examples
    --------
    >>> from pyemv import ac
    >>> sk_ac = bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB")
    >>> data = bytes.fromhex(
    ...     "000000002000" + # 9F02
    ...     "000000000000" + # 9F03
    ...     "0124" +         # 9F1A
    ...     "0000008000" +   # 95
    ...     "0124" +         # 5F2A
    ...     "110309" +       # 9A
    ...     "00" +           # 9C
    ...     "3804823E" +     # 9F37
    ...     "5800" +         # 82
    ...     "0001")          # 9F36
    >>> arqc = ac.generate_ac(sk_ac, data)
    >>> arqc.hex().upper()
    '3B76CF10FECD8789'
    """
    if len(sk_ac) != 16:
        raise ValueError("Session Key must be a double length DES key")

    if padding_type is None:
        padding_type = PaddingType.EMV

    if not isinstance(padding_type, PaddingType):
        raise TypeError(
            "Padding type must be PaddingType Enum, "
            f"not {padding_type.__class__.__name__}"
        )

    return _mac_iso9797_3(sk_ac[:8], sk_ac[-8:], data, padding_type.value, length)


def generate_arpc_1(sk_ac: bytes, arqc: bytes, arpc_rc: bytes) -> bytes:
    r"""Generate Authorisation Response Cryptogram (ARPC) using method 1.
    Method for the generation of an 8-byte ARPC consists of applying
    the Triple-DES algorithm to:

        - 8-byte binary ARQC
        - 2-byte binary Authorisation Response Code (ARPC-RC)

    Parameters
    ----------
    sk_ac : bytes
        Binary ICC Session Key for AC. Has to be a valid DES key.
    arqc : bytes
        Binary 8-byte Authorisation Request Cryptogram (ARQC).
    arpc_rc : bytes
        Binary 2-byte ARPC response code.

    Returns
    -------
    arpc : bytes
        Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
        The resulting issuer authentication data (tag 91) is:

            91 || Len || ARPC || ARPC-RC

    Raises
    ------
    ValueError
        Session Key must be a double length DES key
        ARQC must be 8 bytes long
        ARPC-RC must be 2 bytes long

    Notes
    -----
    After the host has validated the cryptogram sent by the terminal/reader
    in the authorisation request it will generate an Authorisation Response
    cryptogram (ARPC) which will then be returned in the response message
    along with the ARPC response code and ultimately validated by the card.

    For further details see also:
        - EMV 4.3 Book 2 Section 8.2 Issuer Authentication
        - EMV 4.3 Book 2 Section 8.2.1 ARPC Method 1

    Examples
    --------
    >>> from pyemv import ac
    >>> sk_ac = bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB")
    >>> arqc = bytes.fromhex("1234567890ABCDEF")
    >>> arpc_rc = bytes.fromhex("0000")
    >>> arpc = ac.generate_arpc_1(sk_ac, arqc, arpc_rc)
    >>> arpc.hex().upper()
    'F5E6F44147E2F1B0'
    """
    if len(sk_ac) != 16:
        raise ValueError("Session Key must be a double length DES key")

    if len(arqc) != 8:
        raise ValueError("ARQC must be 8 bytes long")

    if len(arpc_rc) != 2:
        raise ValueError("ARPC-RC must be 2 bytes long")

    return _encrypt_tdes_cbc(
        sk_ac, b"\x00\x00\x00\x00\x00\x00\x00\x00", _xor(arpc_rc + (b"\00" * 6), arqc)
    )


def generate_arpc_2(
    sk_ac: bytes,
    arqc: bytes,
    csu: bytes,
    prop_auth_data: _typing.Optional[bytes] = None,
) -> bytes:
    r"""Generate Authorisation Response Cryptogram (ARPC) using method 2.
    Method for the generation of a 4-byte ARPC consists of applying
    ISO/IEC 9797-1 MAC algorithm 3 to:

        - 8-byte binary ARQC
        - 4-byte binary Card Status Update (CSU)
        - 0-8 byte binary Proprietary Authentication Data

    Parameters
    ----------
    sk_ac : bytes
        Binary ICC Session Key for AC. Has to be a valid DES key.
    arqc : bytes
        Binary 8-byte Authorisation Request Cryptogram (ARQC).
    csu : bytes
        Binary 4-byte Card Status Update (CSU).
    prop_auth_data : bytes, optional
        Binary 0-8 byte Proprietary Authentication Data.

    Returns
    -------
    arpc : bytes
        Returns binary 4-byte Authorisation Response Cryptogram (ARPC).
        The resulting issuer authentication data (tag 91) is:

            91 || Len || ARPC || CSU || { Proprietary Authentication Data }

    Raises
    ------
    ValueError
        Session Key must be a double length DES key
        ARQC must be 8 bytes long
        CSU must be 4 bytes long
        Proprietary Authentication Data must be 0-8 bytes long

    Notes
    -----
    After the host has validated the cryptogram sent by the terminal/reader
    in the authorisation request it will generate an Authorisation Response
    cryptogram (ARPC) which will then be returned in the response message
    along with the ARPC response code and ultimately validated by the card.

    For further details see also:
        - EMV 4.3 Book 2 Section 8.2 Issuer Authentication
        - EMV 4.3 Book 2 Section 8.2.2 ARPC Method 2
        - EMV 4.3 Book 2 Annex A 1.2.1

    Examples
    --------
    >>> from pyemv import ac
    >>> sk_ac = bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB")
    >>> arqc = bytes.fromhex("1234567890ABCDEF")
    >>> csu = bytes.fromhex("00000000")
    >>> arpc = ac.generate_arpc_2(sk_ac, arqc, csu)
    >>> arpc.hex().upper()
    '9308BEBC'
    """
    if len(sk_ac) != 16:
        raise ValueError("Session Key must be a double length DES key")

    if prop_auth_data is None:
        prop_auth_data = b""

    if len(arqc) != 8:
        raise ValueError("ARQC must be 8 bytes long")

    if len(csu) != 4:
        raise ValueError("CSU must be 4 bytes long")

    if len(prop_auth_data) > 8:
        raise ValueError("Proprietary Authentication Data must be 0-8 bytes long")

    return _mac_iso9797_3(sk_ac[:8], sk_ac[-8:], arqc + csu + prop_auth_data, 2, 4)



import pytest

from pyemv import ac


def test_generate_ac_exception() -> None:
    # SK < 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        ac.generate_ac(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            data=bytes.fromhex("12345678901214"),
        )

    # SK > 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        ac.generate_ac(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            data=bytes.fromhex("12345678901214"),
        )

    # Invalid padding type
    with pytest.raises(
        TypeError,
        match="Padding type must be PaddingType Enum, not dict",
    ):
        ac.generate_ac(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            data=bytes.fromhex("12345678901214"),
            padding_type={},  # type: ignore
        )


def test_generate_arpc_1_exception() -> None:
    # SK < 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        ac.generate_arpc_1(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            arqc=bytes.fromhex("12345678"),
            arpc_rc=bytes.fromhex("0000"),
        )

    # SK > 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        ac.generate_arpc_1(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            arqc=bytes.fromhex("12345678"),
            arpc_rc=bytes.fromhex("0000"),
        )

    # ARQC < 8 bytes
    with pytest.raises(
        ValueError,
        match="ARQC must be 8 bytes long",
    ):
        ac.generate_arpc_1(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("12345678"),
            arpc_rc=bytes.fromhex("0000"),
        )

    # ARQC > 16 bytes
    with pytest.raises(
        ValueError,
        match="ARQC must be 8 bytes long",
    ):
        ac.generate_arpc_1(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890ABCDEF12"),
            arpc_rc=bytes.fromhex("0000"),
        )

    # ARPC-RC < 2 bytes
    with pytest.raises(
        ValueError,
        match="ARPC-RC must be 2 bytes long",
    ):
        ac.generate_arpc_1(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890123456"),
            arpc_rc=bytes.fromhex("00"),
        )

    # ARPC-RC > 2 bytes
    with pytest.raises(
        ValueError,
        match="ARPC-RC must be 2 bytes long",
    ):
        ac.generate_arpc_1(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890123456"),
            arpc_rc=bytes.fromhex("001122"),
        )


def test_generate_arpc_2_exception() -> None:
    # SK < 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            arqc=bytes.fromhex("12345678901214"),
            csu=bytes.fromhex("12345678"),
            prop_auth_data=bytes.fromhex("1234567890123456"),
        )

    # SK > 16 bytes
    with pytest.raises(
        ValueError,
        match="Session Key must be a double length DES key",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            arqc=bytes.fromhex("12345678901214"),
            csu=bytes.fromhex("12345678"),
            prop_auth_data=bytes.fromhex("1234567890123456"),
        )

    # ARQC < 8 bytes
    with pytest.raises(
        ValueError,
        match="ARQC must be 8 bytes long",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("12345678901214"),
            csu=bytes.fromhex("12345678"),
            prop_auth_data=bytes.fromhex("1234567890123456"),
        )

    # ARQC > 16 bytes
    with pytest.raises(
        ValueError,
        match="ARQC must be 8 bytes long",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890ABCDEF12"),
            csu=bytes.fromhex("12345678"),
            prop_auth_data=bytes.fromhex("1234567890123456"),
        )

    # CSU < 4 bytes
    with pytest.raises(
        ValueError,
        match="CSU must be 4 bytes long",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890121456"),
            csu=bytes.fromhex("123456"),
            prop_auth_data=bytes.fromhex("1234567890123456"),
        )

    # CSU > 4 bytes
    with pytest.raises(
        ValueError,
        match="CSU must be 4 bytes long",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890121456"),
            csu=bytes.fromhex("1234567890"),
            prop_auth_data=bytes.fromhex("1234567890123456"),
        )

    # PAD > 8 bytes
    with pytest.raises(
        ValueError,
        match="Proprietary Authentication Data must be 0-8 bytes long",
    ):
        ac.generate_arpc_2(
            sk_ac=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            arqc=bytes.fromhex("1234567890121456"),
            csu=bytes.fromhex("12345678"),
            prop_auth_data=bytes.fromhex("123456789012345678"),
        )



import pytest

from pyemv import kd


def test_derive_common_sk_exception() -> None:
    # ISS MK < 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key must be a double length DES key",
    ):
        kd.derive_common_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAA"),
            r=bytes.fromhex("FFFFFFFFFFFFFFFF"),
        )

    # ISS MK > 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key must be a double length DES key",
    ):
        kd.derive_common_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            r=bytes.fromhex("FFFFFFFFFFFFFFFF"),
        )

    # R < 8 bytes
    with pytest.raises(
        ValueError,
        match="Diversification value must be 8 bytes long",
    ):
        kd.derive_common_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            r=bytes.fromhex("FFFFFFFFFFFFFF"),
        )

    # R > 8 bytes
    with pytest.raises(
        ValueError,
        match="Diversification value must be 8 bytes long",
    ):
        kd.derive_common_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            r=bytes.fromhex("FFFFFFFFFFFFFFFFFF"),
        )


def test_derive_visa_sm_sk_exception() -> None:
    # ISS MK < 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key must be a double length DES key",
    ):
        kd.derive_visa_sm_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAA"), atc=bytes.fromhex("FFFF")
        )

    # ISS MK > 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key must be a double length DES key",
    ):
        kd.derive_visa_sm_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            atc=bytes.fromhex("FFFF"),
        )

    # ATC < 2 bytes
    with pytest.raises(
        ValueError,
        match="ATC value must be 2 bytes long",
    ):
        kd.derive_visa_sm_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FF"),
        )

    # ATC > 2 bytes
    with pytest.raises(
        ValueError,
        match="ATC value must be 2 bytes long",
    ):
        kd.derive_visa_sm_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FFFFFF"),
        )


def test_derive_emv2000_tree_sk_exception() -> None:
    # ISS MK < 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key must be a double length DES key",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAA"), atc=bytes.fromhex("FFFF")
        )

    # ISS MK > 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC Master Key must be a double length DES key",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCC"),
            atc=bytes.fromhex("FFFF"),
        )

    # ATC < 2 bytes
    with pytest.raises(
        ValueError,
        match="ATC value must be 2 bytes long",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FF"),
        )

    # ATC > 2 bytes
    with pytest.raises(
        ValueError,
        match="ATC value must be 2 bytes long",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FFFFFF"),
        )

    # IV < 16 bytes
    with pytest.raises(
        ValueError,
        match="Initialization vector value must be 16 bytes long",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FFFF"),
            iv=bytes.fromhex("000000000000000000000000000000"),
        )

    # IV > 16 bytes
    with pytest.raises(
        ValueError,
        match="Initialization vector value must be 16 bytes long",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FFFF"),
            iv=bytes.fromhex("0000000000000000000000000000000000"),
        )

    # Branch factor and tree height must be able to generate enough session
    # keys to cover every possible ATC value
    with pytest.raises(
        ValueError,
        match="Number of possible session keys must exceed maximum ATC value",
    ):
        kd.derive_emv2000_tree_sk(
            icc_mk=bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            atc=bytes.fromhex("FFFF"),
            branch_factor=1,
            height=1,
        )



from typing import Union
import pytest
from pyemv import ac, kd, sm, tools


@pytest.mark.parametrize(
    ["pan", "psn"],
    [
        (b"12345678901234567", b"45"),
        (b"12345678901234567", "45"),
        ("12345678901234567", b"45"),
        ("12345678901234567", "45"),
    ],
)
def test_derive_icc_mk_a_psn(pan: Union[bytes, str], psn: Union[bytes, str]) -> None:
    """
    Verify ICC MK derivation method A with non-zero PSN.

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using ICC Master Key
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "FF08"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "DF82"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "19C1FBC83EBDC0D5"

    # ARPC Method 1 using ICC Master Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(icc_mk, arqc, arpc_rc)
    assert arpc.hex().upper() == "78A372523FA35A03"


@pytest.mark.parametrize(
    "pan",
    [
        b"12345678901234567",
        "12345678901234567",
    ],
)
def test_derive_icc_mk_a_no_psn(pan: Union[bytes, str]) -> None:
    """
    Verify ICC MK derivation method A with a zero PSN.

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using ICC Master Key
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan)

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "3F4F"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "8698A7319324FD93"

    # ARPC Method 1 using ICC Master Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(icc_mk, arqc, arpc_rc)
    assert arpc.hex().upper() == "BEA11C8F4A47EF6F"


def test_derive_common_sk() -> None:
    """
    Verify common session key derivation using algorithm
    type where both ARQC and ARPC are verified using derived
    session key.

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key Derivation
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "12345678901234567"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "FF08"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "DF82"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "19C1FBC83EBDC0D5"

    # ARPC Method 1 using Session Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(sk_ac, arqc, arpc_rc)
    assert arpc.hex().upper() == "C3620580668E5B65"


def test_derive_icc_mk_b_pan16() -> None:
    """
    Verify ICC MK derivation method B using incompatible
    PAN length. Method B is applicable only if PAN is
    17-19 digits long.

    Master Key Derivation = Option B
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key Derivation
    ARPC Method = 2
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "1234567890123456"
    icc_mk = kd.derive_icc_mk_b(iss_mk, pan)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "BAB0"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "BC19"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "0CE77D211CB5459A"

    # ARPC Method 2 using Session Key
    csu = bytes.fromhex("00000000")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu)
    assert arpc.hex().upper() == "8CD9AA5D"


@pytest.mark.parametrize(
    ["pan", "psn"],
    [
        (b"12345678901234567", b"45"),
        (b"12345678901234567", "45"),
        ("12345678901234567", b"45"),
        ("12345678901234567", "45"),
    ],
)
def test_derive_icc_mk_b_pan17_psn(
    pan: Union[bytes, str], psn: Union[bytes, str]
) -> None:
    """
    Verify ICC MK derivation method B using 17 digit PAN.

    Master Key Derivation = Option B
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key Derivation
    ARPC Method = 2
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    icc_mk = kd.derive_icc_mk_b(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "0BAF"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "4262"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "5760EE07B4FA65D1"

    # ARPC Method 2 using Session Key
    csu = bytes.fromhex("00000000")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu)
    assert arpc.hex().upper() == "106B81D9"


@pytest.mark.parametrize(
    "pan",
    [
        b"12345678901234567",
        "12345678901234567",
    ],
)
def test_derive_icc_mk_b_pan17_no_psn(pan: Union[bytes, str]) -> None:
    """
    Verify ICC MK derivation method B using 17 digit PAN.

    Master Key Derivation = Option B
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key Derivation
    ARPC Method = 2
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    icc_mk = kd.derive_icc_mk_b(iss_mk, pan)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "4626"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "7F36"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "0BAA251EA8989442"

    # ARPC Method 2 using Session Key
    csu = bytes.fromhex("00000000")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu)
    assert arpc.hex().upper() == "C1C41F3A"


def test_derive_icc_mk_b_pan18() -> None:
    """
    Verify ICC MK derivation method B using 18 digit PAN.

    Master Key Derivation = Option B
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key Derivation
    ARPC Method = 2
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "123456789012345679"
    psn = "00"
    icc_mk = kd.derive_icc_mk_b(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "C2F3"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "7C35"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "DC95BCE3EBBE0296"

    # ARPC Method 2 using Session Key
    csu = bytes.fromhex("00000000")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu)
    assert arpc.hex().upper() == "ECCA0C4B"


def test_derive_icc_mk_b_sha_pad() -> None:
    """
    Verify ICC MK derivation method B where the algorithm
    is forced to convert sha digest letters into numbers.

    Master Key Derivation = Option B
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key Derivation
    ARPC Method = 2

    Note: in this test sha1 does not produce 16 digits.
    Use decimalisation table to convert hexchars to digits.
        e -> 4
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "000000000000000005"
    psn = "23"
    icc_mk = kd.derive_icc_mk_b(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "DD73"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("001C000000000000")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "04F8"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex(
        "00000000400000000000000001248000048000012"
        "41911050152BF45851800001C06011203A0B800"
    )
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "8CAD6F2489C640B1"

    # ARPC Method 2 using Session Key
    arqc = bytes.fromhex("8CAD6F2489C640B1")
    csu = bytes.fromhex("00000000")
    prop_auth_data = bytes.fromhex("12345678")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu, prop_auth_data)
    assert arpc.hex().upper() == "E39F1876"


def test_derive_visa_sm_sk() -> None:
    """
    Verify visa session key derivation for secure messaging.
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("FEDCBA98765432100123456789ABCDEF")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "7B83"

    # Derive ICC master key
    pan = "1234567890123456"
    icc_mk = kd.derive_icc_mk_b(iss_mk, pan)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "F010"

    # Verify AC session key
    atc = bytes.fromhex("001C")
    arqc = bytes.fromhex("29CCA15AE665FA2E")
    sk_sm = kd.derive_visa_sm_sk(icc_mk, atc)
    assert tools.key_check_digits(sk_sm, 2).hex().upper() == "DB10"

    # Command MAC generation using Session Key
    command_header = bytes.fromhex("8418000008")
    mac = sm.generate_command_mac(sk_sm, command_header + atc + arqc)

    assert mac.hex().upper() == "DB56BA60087CEFD3"


# fmt: off
@pytest.mark.parametrize(
    ["atc", "height", "branch", "iv", "sk"],
    [
        # Small ATC
        ("0000",  8, 4, "00000000000000000000000000000000", "80BF3145AECB948A5201DAB9EF469875"),
        ("0000",  8, 4, "F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0", "DC31BC40B3AB8F73F2E3DA5EBF9DE59B"),
        ("0000", 16, 2, "00000000000000000000000000000000", "86AD9732D5A192A749ADCE31919DAE8A"),
        ("0000", 16, 2, "F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0", "7C57A15261202A101C6DA29B16672A3E"),
        # Medium ATC
        ("00FF",  8, 4, "00000000000000000000000000000000", "981F9D208AE6132CBC8A7C97F42A5445"),
        ("00FF",  8, 4, "F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0", "D579C8A4927589017FA1739B2C853BAD"),
        ("00FF", 16, 2, "00000000000000000000000000000000", "1F54F492EF104580B0F4B0F8910D0438"),
        ("00FF", 16, 2, "F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0", "2086F81AAB731FF8CED9BF6DA7191CA2"),
        # Large ATC
        ("FFFF",  8, 4, "00000000000000000000000000000000", "D93D7A6B70F867615D4AAB79BF235B62"),
        ("FFFF",  8, 4, "F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0", "A10EA78F2F1ACE203B3B402F6EE6CB29"),
        ("FFFF", 16, 2, "00000000000000000000000000000000", "F83E499EF786C20BC41310A176E3F84F"),
        ("FFFF", 16, 2, "F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0", "73E3B0A1406E5E4908D5804C4970E520"),
    ],
)
# fmt: on
def test_derive_emv2000_tree_sk(
    atc: str,
    height: int,
    branch: int,
    iv: str,
    sk: str,
) -> None:
    assert bytes.fromhex(sk) == kd.derive_emv2000_tree_sk(
        bytes.fromhex("0123456789ABCDEF0123456789ABCDEF"),
        bytes.fromhex(atc),
        height,
        branch,
        bytes.fromhex(iv),
    )



from setuptools import find_packages, setup

from pyemv import __version__

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Security :: Cryptography",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

if __name__ == "__main__":

    with open("README.rst", "r", encoding="utf-8") as f:
        readme = f.read()

    setup(
        name="pyemv",
        version=__version__,
        author="Konstantin Novichikhin",
        author_email="konstantin.novichikhin@gmail.com",
        description="A Python package for EMV cryptography in payment systems",
        long_description=readme,
        long_description_content_type="text/x-rst",
        license="MIT",
        url="https://github.com/knovichikhin/pyemv",
        packages=find_packages(exclude=["tests"]),
        package_data={"pyemv": ["py.typed"]},
        zip_safe=False,
        install_requires=[
            "cryptography >= 2.8",
        ],
        classifiers=classifiers,
        python_requires=">=3.6",
        keywords="emv arqc arpc tc aac smi smc cvc3 tlv",
    )
